// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2017 and onwards Google, Inc.

#ifndef BAUMWELCH_LOG_MATH_H_
#define BAUMWELCH_LOG_MATH_H_

// Fast semiring mathematics on POD float types:
//
// * LogTimes is addition (cf. Log and Tropical).
// * LogDivide is subtraction (cf. Log and Tropical).
// * LogPlus is exp, addition, and log (cf. Log).
// * LogApproxEqual uses an addition-based comparison (cf. ApproxEqual).
// * LogUniformGenerator samples uniformly from \log (0, 1].

#include <cmath>
#include <cstdlib>

#include <fst/fstlib.h>

namespace fst {
namespace internal {

template <class T>
inline T LogExp(T t) {
  return log(1.0F + exp(-t));
}

template <class T>
inline T LogPlus(T x, T y) {
  if (x == FloatLimits<T>::PosInfinity()) {
    return y;
  } else if (y == FloatLimits<T>::PosInfinity()) {
    return x;
  } else {
    return x > y ? y - LogExp(x - y) : x - LogExp(y - x);
  }
}

template <class T>
inline T LogTimes(T x, T y) {
  return x + y;
}

// Three-way version of the same.
template <class T>
inline T LogTimes(T x, T y, T z) {
  return x + y + z;
}

template <class T>
inline T LogDivide(T x, T y) {
  if (y == FloatLimits<T>::PosInfinity()) {
    LOG(WARNING) << "LogDivide: unexpected Zero denominator";
    return x == FloatLimits<T>::PosInfinity() ? FloatLimits<T>::NumberBad() :
        FloatLimits<T>::NegInfinity();
  } else if (x == FloatLimits<T>::PosInfinity()) {
    return x;
  } else {
    return x - y;
  }
}

template <class T>
inline bool LogApproxEqual(T x, T y, float delta) {
  return x <= y + delta && y <= x + delta;
}

// Random weight generation in the (real) interval (0, 1].
template <class T>
class LogUniformGenerator {
 public:
  LogUniformGenerator() {}

  // This is inclusive on the lower bound but exclusive on the upper bound. We
  // want the opposite (inclusive on the upper bound by exclusive on the lower
  // bound) so we map any sample right at the lower bound onto the upper bound.
  T operator()() const {
    const auto sample = static_cast<T>(rand()) / RAND_MAX;  // NOLINT
    return -(sample == 0.0 ? 0.0 /* i.e., log(1) */ : log(sample));
  }
};

}  // namespace internal
}  // namespace fst

#endif  //  BAUMWELCH_LOG_MATH_H_

